/*
The software in this package is distributed under the GNU General
Public License version 2 (with a special exception described below).

A copy of GNU General Public License (GPL) is included in this distribution,
in the file COPYING.GPL.

As a special exception, if other files instantiate templates or use macros
or inline functions from this file, or you compile this file and link it
with other works to produce a work based on this file, this file
does not by itself cause the resulting work to be covered
by the GNU General Public License.

However the source code for this file must still be made available
in accordance with section (3) of the GNU General Public License.

This exception does not invalidate any other reasons why a work based
on this file might be covered by the GNU General Public License.
*/
#ifndef TCP_HEADER_H
#define TCP_HEADER_H

#include <stdint.h>

#include <istream>
#include <ostream>

#include <boost/asio.hpp>

#include "host/messagepayload.h"

//-----------------------------------------------------------------------------
// TcpHeader
//-----------------------------------------------------------------------------

/**
 * @brief This class represents the TCP Segment Header.
 *
 * The wire format of a TCP header is:
 *
 * @code
 * 0        4       8              16               24           31
 * +-------------------------------+------------------------------+      ---
 * |                               |                              |       ^
 * |         source port           |     destination port         |       |
 * |                               |                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |       |
 * |                        sequence number                       |       |
 * |                                                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |       |
 * |              acknowledgment number (if ACK set)              |   20 bytes
 * |                                                              |       |
 * +-------+-------+-+-+-+-+-+-+-+-+------------------------------+       |
 * |  TCP  | reser-|C|E|U|A|P|R|S|F|                              |       |
 * | header| ved   |W|C|R|C|S|S|Y|I|        window size           |       |
 * |length |       |R|E|G|K|H|T|N|N|                              |       |
 * +---------------+-+-+-+-+-+-+-+-+------------------------------+       |
 * |                               |                              |       |
 * |          checksum             | urgent pointer (if URG set)  |       |
 * |                               |                              |       v
 * +-----------------------------------------------+--------------+      ---
 * |                                               |              |       ^
 * |                                               |              |       |
 * |          options (if Data Offset > 5)         |   padding    /    0 - 40
 * |                                               |              /     bytes
 * |                                               |              |       |
 * |                                               |              |       v
 * +-----------------------------------------------+--------------+      ---
 * @endcode
 */
class TcpHeader
{
public:
    TcpHeader();

    uint16_t get_source_port() const;
    uint16_t get_destination_port() const;
    uint32_t get_sequence_number() const;
    uint32_t get_acknowledgment_number() const;
    uint8_t get_header_length() const;
    bool get_congestion_window_reduced() const;
    bool get_ecn_echo() const;
    bool get_urgent() const;
    bool get_acknowledgment() const;
    bool get_push() const;
    bool get_reset() const;
    bool get_synchronize() const;
    bool get_finish() const;
    uint16_t get_window_size() const;
    uint16_t get_checksum() const;

    uint16_t calculate_tcp_checksum(
            const boost::asio::ip::address_v4::bytes_type &src_addr,
            const boost::asio::ip::address_v4::bytes_type &dest_addr,
            const uint8_t *tcp_payload_data,
            const uint16_t tcp_payload_size_in_bytes
    );
    uint16_t calculate_tcp_checksum(
            const boost::asio::ip::address_v6::bytes_type &src_addr,
            const boost::asio::ip::address_v6::bytes_type &dest_addr,
            const uint8_t *tcp_payload_data,
            const uint16_t tcp_payload_size_in_bytes
    );

    void set_source_port( const uint16_t port );
    void set_destination_port( const uint16_t port );
    void set_sequence_number( const uint32_t seq_num );
    void set_acknowledgment_number( const uint32_t ack_num );
    void set_header_length( const uint8_t offset );
    void set_congestion_window_reduced( bool bit );
    void set_ecn_echo( bool bit );
    void set_urgent( bool bit );
    void set_acknowledgment( bool bit );
    void set_push( bool bit );
    void set_reset( bool bit );
    void set_synchronize( bool bit );
    void set_finish( bool bit );
    void set_window_size( const uint16_t wnd_size );
    void set_checksum( const uint16_t sum );

    friend std::istream& operator>>(
            std::istream &is,
            TcpHeader &header
    );
    friend std::ostream& operator<<(
            std::ostream &os,
            const TcpHeader &header
    );

private:
    uint16_t calculate_checksum(
            const uint16_t *word_array,
            const std::size_t size
    ) const;

    MessagePayload Payload;
};

#endif // TCP_HEADER_H
